@@ -3,7 +3,7 @@
  *				back to source text
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/utils/adt/ruleutils.c,v 1.77.2.1 2001/07/31 18:39:12 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/ruleutils.c,v 1.89.2.1 2002/06/15 18:38:10 tgl Exp $
  *
  *	  This software is copyrighted by Jan Wieck - Hamburg.
  *
@@ -40,10 +40,12 @@
 #include <unistd.h>
 #include <fcntl.h>
 
+#include "catalog/heap.h"
+#include "catalog/index.h"
 #include "catalog/pg_index.h"
+#include "catalog/pg_opclass.h"
 #include "catalog/pg_operator.h"
 #include "catalog/pg_shadow.h"
-#include "commands/view.h"
 #include "executor/spi.h"
 #include "lib/stringinfo.h"
 #include "optimizer/clauses.h"
@@ -52,7 +54,7 @@
 #include "parser/parse_expr.h"
 #include "parser/parsetree.h"
 #include "rewrite/rewriteManip.h"
-#include "utils/fmgroids.h"
+#include "rewrite/rewriteSupport.h"
 #include "utils/lsyscache.h"
 
 
@@ -82,7 +84,7 @@ typedef struct
 	List	   *rtable;			/* List of RangeTblEntry nodes */
 	List	   *namespace;		/* List of joinlist items (RangeTblRef and
 								 * JoinExpr nodes) */
-}			deparse_namespace;
+} deparse_namespace;
 
 
 /* ----------
@@ -94,10 +96,6 @@ static void *plan_getrule = NULL;
 static char *query_getrule = "SELECT * FROM pg_rewrite WHERE rulename = $1";
 static void *plan_getview = NULL;
 static char *query_getview = "SELECT * FROM pg_rewrite WHERE rulename = $1";
-static void *plan_getam = NULL;
-static char *query_getam = "SELECT * FROM pg_am WHERE oid = $1";
-static void *plan_getopclass = NULL;
-static char *query_getopclass = "SELECT * FROM pg_opclass WHERE oid = $1";
 
 
 /* ----------
@@ -118,10 +116,10 @@ static void get_delete_query_def(Query *query, deparse_context *context);
 static void get_utility_query_def(Query *query, deparse_context *context);
 static void get_basic_select_query(Query *query, deparse_context *context);
 static void get_setop_query(Node *setOp, Query *query,
-				deparse_context *context, bool toplevel);
+				deparse_context *context);
 static void get_rule_sortgroupclause(SortClause *srt, List *tlist,
-									 bool force_colno,
-									 deparse_context *context);
+						 bool force_colno,
+						 deparse_context *context);
 static void get_names_for_var(Var *var, deparse_context *context,
 				  char **refname, char **attname);
 static bool get_alias_for_case(CaseExpr *caseexpr, deparse_context *context,
@@ -132,12 +130,15 @@ static bool find_alias_in_namespace(Node *nsnode, Node *expr,
 static bool phony_equal(Node *expr1, Node *expr2, int levelsup);
 static void get_rule_expr(Node *node, deparse_context *context);
 static void get_func_expr(Expr *expr, deparse_context *context);
+static Node *strip_type_coercion(Node *expr, Oid resultType);
 static void get_tle_expr(TargetEntry *tle, deparse_context *context);
 static void get_const_expr(Const *constval, deparse_context *context);
 static void get_sublink_expr(Node *node, deparse_context *context);
 static void get_from_clause(Query *query, deparse_context *context);
 static void get_from_clause_item(Node *jtnode, Query *query,
 					 deparse_context *context);
+static void get_opclass_name(Oid opclass, Oid actual_datatype,
+				 StringInfo buf);
 static bool tleIsArrayAssign(TargetEntry *tle);
 static char *quote_identifier(char *ident);
 static char *get_relation_name(Oid relid);
@@ -177,9 +178,9 @@ pg_get_ruledef(PG_FUNCTION_ARGS)
 		elog(ERROR, "get_ruledef: cannot connect to SPI manager");
 
 	/*
-	 * On the first call prepare the plan to lookup pg_proc. We read
-	 * pg_proc over the SPI manager instead of using the syscache to be
-	 * checked for read access on pg_proc.
+	 * On the first call prepare the plan to lookup pg_rewrite. We read
+	 * pg_rewrite over the SPI manager instead of using the syscache to be
+	 * checked for read access on pg_rewrite.
 	 */
 	if (plan_getrule == NULL)
 	{
@@ -270,9 +271,9 @@ pg_get_viewdef(PG_FUNCTION_ARGS)
 		elog(ERROR, "get_viewdef: cannot connect to SPI manager");
 
 	/*
-	 * On the first call prepare the plan to lookup pg_proc. We read
-	 * pg_proc over the SPI manager instead of using the syscache to be
-	 * checked for read access on pg_proc.
+	 * On the first call prepare the plan to lookup pg_rewrite. We read
+	 * pg_rewrite over the SPI manager instead of using the syscache to be
+	 * checked for read access on pg_rewrite.
 	 */
 	if (plan_getview == NULL)
 	{
@@ -301,7 +302,6 @@ pg_get_viewdef(PG_FUNCTION_ARGS)
 		appendStringInfo(&buf, "Not a view");
 	else
 	{
-
 		/*
 		 * Get the rules definition and put it into executors memory
 		 */
@@ -341,48 +341,16 @@ pg_get_indexdef(PG_FUNCTION_ARGS)
 	HeapTuple	ht_idx;
 	HeapTuple	ht_idxrel;
 	HeapTuple	ht_indrel;
-	HeapTuple	spi_tup;
-	TupleDesc	spi_ttc;
-	int			spi_fno;
 	Form_pg_index idxrec;
 	Form_pg_class idxrelrec;
 	Form_pg_class indrelrec;
-	Datum		spi_args[1];
-	char		spi_nulls[2];
-	int			spirc;
+	Form_pg_am	amrec;
 	int			len;
 	int			keyno;
 	StringInfoData buf;
 	StringInfoData keybuf;
 	char	   *sep;
 
-	/*
-	 * Connect to SPI manager
-	 */
-	if (SPI_connect() != SPI_OK_CONNECT)
-		elog(ERROR, "get_indexdef: cannot connect to SPI manager");
-
-	/*
-	 * On the first call prepare the plans to lookup pg_am and pg_opclass.
-	 */
-	if (plan_getam == NULL)
-	{
-		Oid			argtypes[1];
-		void	   *plan;
-
-		argtypes[0] = OIDOID;
-		plan = SPI_prepare(query_getam, 1, argtypes);
-		if (plan == NULL)
-			elog(ERROR, "SPI_prepare() failed for \"%s\"", query_getam);
-		plan_getam = SPI_saveplan(plan);
-
-		argtypes[0] = OIDOID;
-		plan = SPI_prepare(query_getopclass, 1, argtypes);
-		if (plan == NULL)
-			elog(ERROR, "SPI_prepare() failed for \"%s\"", query_getopclass);
-		plan_getopclass = SPI_saveplan(plan);
-	}
-
 	/*
 	 * Fetch the pg_index tuple by the Oid of the index
 	 */
@@ -414,21 +382,14 @@ pg_get_indexdef(PG_FUNCTION_ARGS)
 	indrelrec = (Form_pg_class) GETSTRUCT(ht_indrel);
 
 	/*
-	 * Get the am name for the index relation
+	 * Fetch the pg_am tuple of the index' access method
+	 *
+	 * There is no syscache for this, so use index.c subroutine.
 	 */
-	spi_args[0] = ObjectIdGetDatum(idxrelrec->relam);
-	spi_nulls[0] = ' ';
-	spi_nulls[1] = '\0';
-	spirc = SPI_execp(plan_getam, spi_args, spi_nulls, 1);
-	if (spirc != SPI_OK_SELECT)
-		elog(ERROR, "failed to get pg_am tuple for index %s",
-			 NameStr(idxrelrec->relname));
-	if (SPI_processed != 1)
-		elog(ERROR, "failed to get pg_am tuple for index %s",
-			 NameStr(idxrelrec->relname));
-	spi_tup = SPI_tuptable->vals[0];
-	spi_ttc = SPI_tuptable->tupdesc;
-	spi_fno = SPI_fnumber(spi_ttc, "amname");
+	amrec = AccessMethodObjectIdGetForm(idxrelrec->relam,
+										CurrentMemoryContext);
+	if (!amrec)
+		elog(ERROR, "lookup for AM %u failed", idxrelrec->relam);
 
 	/*
 	 * Start the index definition
@@ -436,19 +397,20 @@ pg_get_indexdef(PG_FUNCTION_ARGS)
 	initStringInfo(&buf);
 	appendStringInfo(&buf, "CREATE %sINDEX %s ON %s USING %s (",
 					 idxrec->indisunique ? "UNIQUE " : "",
-				  quote_identifier(pstrdup(NameStr(idxrelrec->relname))),
-				  quote_identifier(pstrdup(NameStr(indrelrec->relname))),
-					 quote_identifier(SPI_getvalue(spi_tup, spi_ttc,
-												   spi_fno)));
+					 quote_identifier(NameStr(idxrelrec->relname)),
+					 quote_identifier(NameStr(indrelrec->relname)),
+					 quote_identifier(NameStr(amrec->amname)));
 
 	/*
-	 * Collect the indexed attributes
+	 * Collect the indexed attributes in keybuf
 	 */
 	initStringInfo(&keybuf);
 	sep = "";
 	for (keyno = 0; keyno < INDEX_MAX_KEYS; keyno++)
 	{
-		if (idxrec->indkey[keyno] == InvalidAttrNumber)
+		AttrNumber	attnum = idxrec->indkey[keyno];
+
+		if (attnum == InvalidAttrNumber)
 			break;
 
 		appendStringInfo(&keybuf, sep);
@@ -459,35 +421,22 @@ pg_get_indexdef(PG_FUNCTION_ARGS)
 		 */
 		appendStringInfo(&keybuf, "%s",
 			  quote_identifier(get_relid_attribute_name(idxrec->indrelid,
-												idxrec->indkey[keyno])));
+														attnum)));
 
 		/*
 		 * If not a functional index, add the operator class name
 		 */
 		if (idxrec->indproc == InvalidOid)
-		{
-			spi_args[0] = ObjectIdGetDatum(idxrec->indclass[keyno]);
-			spi_nulls[0] = ' ';
-			spi_nulls[1] = '\0';
-			spirc = SPI_execp(plan_getopclass, spi_args, spi_nulls, 1);
-			if (spirc != SPI_OK_SELECT)
-				elog(ERROR, "failed to get pg_opclass tuple %u", idxrec->indclass[keyno]);
-			if (SPI_processed != 1)
-				elog(ERROR, "failed to get pg_opclass tuple %u", idxrec->indclass[keyno]);
-			spi_tup = SPI_tuptable->vals[0];
-			spi_ttc = SPI_tuptable->tupdesc;
-			spi_fno = SPI_fnumber(spi_ttc, "opcname");
-			appendStringInfo(&keybuf, " %s",
-						  quote_identifier(SPI_getvalue(spi_tup, spi_ttc,
-														spi_fno)));
-		}
+			get_opclass_name(idxrec->indclass[keyno],
+							 get_atttype(idxrec->indrelid, attnum),
+							 &keybuf);
 	}
 
-	/*
-	 * For functional index say 'func (attrs) opclass'
-	 */
 	if (idxrec->indproc != InvalidOid)
 	{
+		/*
+		 * For functional index say 'func (attrs) opclass'
+		 */
 		HeapTuple	proctup;
 		Form_pg_proc procStruct;
 
@@ -498,62 +447,128 @@ pg_get_indexdef(PG_FUNCTION_ARGS)
 			elog(ERROR, "cache lookup for proc %u failed", idxrec->indproc);
 		procStruct = (Form_pg_proc) GETSTRUCT(proctup);
 
-		appendStringInfo(&buf, "%s(%s) ",
-				 quote_identifier(pstrdup(NameStr(procStruct->proname))),
+		appendStringInfo(&buf, "%s(%s)",
+						 quote_identifier(NameStr(procStruct->proname)),
 						 keybuf.data);
+		get_opclass_name(idxrec->indclass[0], procStruct->prorettype, &buf);
 
-		spi_args[0] = ObjectIdGetDatum(idxrec->indclass[0]);
-		spi_nulls[0] = ' ';
-		spi_nulls[1] = '\0';
-		spirc = SPI_execp(plan_getopclass, spi_args, spi_nulls, 1);
-		if (spirc != SPI_OK_SELECT)
-			elog(ERROR, "failed to get pg_opclass tuple %u", idxrec->indclass[0]);
-		if (SPI_processed != 1)
-			elog(ERROR, "failed to get pg_opclass tuple %u", idxrec->indclass[0]);
-		spi_tup = SPI_tuptable->vals[0];
-		spi_ttc = SPI_tuptable->tupdesc;
-		spi_fno = SPI_fnumber(spi_ttc, "opcname");
-		appendStringInfo(&buf, "%s",
-						 quote_identifier(SPI_getvalue(spi_tup, spi_ttc,
-													   spi_fno)));
 		ReleaseSysCache(proctup);
 	}
 	else
-
+	{
 		/*
-		 * For the others say 'attr opclass [, ...]'
+		 * Otherwise say 'attr opclass [, ...]'
 		 */
 		appendStringInfo(&buf, "%s", keybuf.data);
+	}
 
-	/*
-	 * Finish
-	 */
 	appendStringInfo(&buf, ")");
 
 	/*
-	 * Create the result in upper executor memory, and free objects
+	 * If it's a partial index, decompile and append the predicate
+	 */
+	if (VARSIZE(&idxrec->indpred) > VARHDRSZ)
+	{
+		Node	   *node;
+		List	   *context;
+		char	   *exprstr;
+		char	   *str;
+
+		/* Convert TEXT object to C string */
+		exprstr = DatumGetCString(DirectFunctionCall1(textout,
+									 PointerGetDatum(&idxrec->indpred)));
+		/* Convert expression to node tree */
+		node = (Node *) stringToNode(exprstr);
+
+		/*
+		 * If top level is a List, assume it is an implicit-AND structure,
+		 * and convert to explicit AND.  This is needed for partial index
+		 * predicates.
+		 */
+		if (node && IsA(node, List))
+			node = (Node *) make_ands_explicit((List *) node);
+		/* Deparse */
+		context = deparse_context_for(NameStr(indrelrec->relname),
+									  idxrec->indrelid);
+		str = deparse_expression(node, context, false);
+		appendStringInfo(&buf, " WHERE %s", str);
+	}
+
+	/*
+	 * Create the result as a TEXT datum, and free working data
 	 */
 	len = buf.len + VARHDRSZ;
-	indexdef = SPI_palloc(len);
+	indexdef = (text *) palloc(len);
 	VARATT_SIZEP(indexdef) = len;
 	memcpy(VARDATA(indexdef), buf.data, buf.len);
 
 	pfree(buf.data);
 	pfree(keybuf.data);
+	pfree(amrec);
 	ReleaseSysCache(ht_idx);
 	ReleaseSysCache(ht_idxrel);
 	ReleaseSysCache(ht_indrel);
 
-	/*
-	 * Disconnect from SPI manager
-	 */
-	if (SPI_finish() != SPI_OK_FINISH)
-		elog(ERROR, "get_viewdef: SPI_finish() failed");
-
 	PG_RETURN_TEXT_P(indexdef);
 }
 
 
+/* ----------
+ * get_expr			- Decompile an expression tree
+ *
+ * Input: an expression tree in nodeToString form, and a relation OID
+ *
+ * Output: reverse-listed expression
+ *
+ * Currently, the expression can only refer to a single relation, namely
+ * the one specified by the second parameter.  This is sufficient for
+ * partial indexes, column default expressions, etc.
+ * ----------
+ */
+Datum
+pg_get_expr(PG_FUNCTION_ARGS)
+{
+	text	   *expr = PG_GETARG_TEXT_P(0);
+	Oid			relid = PG_GETARG_OID(1);
+	text	   *result;
+	Node	   *node;
+	List	   *context;
+	char	   *exprstr;
+	char	   *relname;
+	char	   *str;
+
+	/* Get the name for the relation */
+	relname = get_rel_name(relid);
+	if (relname == NULL)
+		PG_RETURN_NULL();		/* should we raise an error? */
+
+	/* Convert input TEXT object to C string */
+	exprstr = DatumGetCString(DirectFunctionCall1(textout,
+												  PointerGetDatum(expr)));
+
+	/* Convert expression to node tree */
+	node = (Node *) stringToNode(exprstr);
+
+	/*
+	 * If top level is a List, assume it is an implicit-AND structure, and
+	 * convert to explicit AND.  This is needed for partial index
+	 * predicates.
+	 */
+	if (node && IsA(node, List))
+		node = (Node *) make_ands_explicit((List *) node);
+
+	/* Deparse */
+	context = deparse_context_for(relname, relid);
+	str = deparse_expression(node, context, false);
+
+	/* Pass the result back as TEXT */
+	result = DatumGetTextP(DirectFunctionCall1(textin,
+											   CStringGetDatum(str)));
+
+	PG_RETURN_TEXT_P(result);
+}
+
+
 /* ----------
  * get_userbyid			- Get a user name by usesysid and
  *				  fallback to 'unknown (UID=n)'
@@ -754,8 +769,21 @@ make_ruledef(StringInfo buf, HeapTuple ruletup, TupleDesc rulettc)
 		appendStringInfo(buf, " WHERE ");
 
 		qual = stringToNode(ev_qual);
+
+		/*
+		 * We need to make a context for recognizing any Vars in the qual
+		 * (which can only be references to OLD and NEW).  Use the rtable
+		 * of the first query in the action list for this purpose.
+		 */
 		query = (Query *) lfirst(actions);
 
+		/*
+		 * If the action is INSERT...SELECT, OLD/NEW have been pushed
+		 * down into the SELECT, and that's what we need to look at.
+		 * (Ugly kluge ... try to fix this when we redesign querytrees.)
+		 */
+		query = getInsertSelectQuery(query, NULL);
+
 		context.buf = buf;
 		context.namespaces = makeList1(&dpns);
 		context.varprefix = (length(query->rtable) != 1);
@@ -938,7 +966,7 @@ get_select_query_def(Query *query, deparse_context *context)
 	 */
 	if (query->setOperations)
 	{
-		get_setop_query(query->setOperations, query, context, true);
+		get_setop_query(query->setOperations, query, context);
 		/* ORDER BY clauses must be simple in this case */
 		force_colno = true;
 	}
@@ -1095,8 +1123,7 @@ get_basic_select_query(Query *query, deparse_context *context)
 }
 
 static void
-get_setop_query(Node *setOp, Query *query, deparse_context *context,
-				bool toplevel)
+get_setop_query(Node *setOp, Query *query, deparse_context *context)
 {
 	StringInfo	buf = context->buf;
 
@@ -1113,33 +1140,29 @@ get_setop_query(Node *setOp, Query *query, deparse_context *context,
 	{
 		SetOperationStmt *op = (SetOperationStmt *) setOp;
 
-		/*
-		 * Must suppress parens at top level of a setop tree because of
-		 * grammar limitations...
-		 */
-		if (!toplevel)
-			appendStringInfo(buf, "(");
-		get_setop_query(op->larg, query, context, false);
+		appendStringInfo(buf, "((");
+		get_setop_query(op->larg, query, context);
 		switch (op->op)
 		{
 			case SETOP_UNION:
-				appendStringInfo(buf, " UNION ");
+				appendStringInfo(buf, ") UNION ");
 				break;
 			case SETOP_INTERSECT:
-				appendStringInfo(buf, " INTERSECT ");
+				appendStringInfo(buf, ") INTERSECT ");
 				break;
 			case SETOP_EXCEPT:
-				appendStringInfo(buf, " EXCEPT ");
+				appendStringInfo(buf, ") EXCEPT ");
 				break;
 			default:
 				elog(ERROR, "get_setop_query: unexpected set op %d",
 					 (int) op->op);
 		}
 		if (op->all)
-			appendStringInfo(buf, "ALL ");
-		get_setop_query(op->rarg, query, context, false);
-		if (!toplevel)
-			appendStringInfo(buf, ")");
+			appendStringInfo(buf, "ALL (");
+		else
+			appendStringInfo(buf, "(");
+		get_setop_query(op->rarg, query, context);
+		appendStringInfo(buf, "))");
 	}
 	else
 	{
@@ -1161,10 +1184,11 @@ get_rule_sortgroupclause(SortClause *srt, List *tlist, bool force_colno,
 
 	tle = get_sortgroupclause_tle(srt, tlist);
 	expr = tle->expr;
+
 	/*
 	 * Use column-number form if requested by caller or if expression is a
-	 * constant --- a constant is ambiguous (and will be misinterpreted
-	 * by findTargetlistEntry()) if we dump it explicitly.
+	 * constant --- a constant is ambiguous (and will be misinterpreted by
+	 * findTargetlistEntry()) if we dump it explicitly.
 	 */
 	if (force_colno || (expr && IsA(expr, Const)))
 	{
@@ -1837,28 +1861,36 @@ get_rule_expr(Node *node, deparse_context *context)
 		case T_FieldSelect:
 			{
 				FieldSelect *fselect = (FieldSelect *) node;
+				Oid			argType = exprType(fselect->arg);
 				HeapTuple	typetup;
 				Form_pg_type typeStruct;
 				Oid			typrelid;
 				char	   *fieldname;
 
-				/* we do NOT parenthesize the arg expression, for now */
-				get_rule_expr(fselect->arg, context);
+				/* lookup arg type and get the field name */
 				typetup = SearchSysCache(TYPEOID,
-								ObjectIdGetDatum(exprType(fselect->arg)),
+										 ObjectIdGetDatum(argType),
 										 0, 0, 0);
 				if (!HeapTupleIsValid(typetup))
 					elog(ERROR, "cache lookup of type %u failed",
-						 exprType(fselect->arg));
+						 argType);
 				typeStruct = (Form_pg_type) GETSTRUCT(typetup);
 				typrelid = typeStruct->typrelid;
 				if (!OidIsValid(typrelid))
 					elog(ERROR, "Argument type %s of FieldSelect is not a tuple type",
-						 NameStr(typeStruct->typname));
+						 format_type_be(argType));
+				ReleaseSysCache(typetup);
 				fieldname = get_relid_attribute_name(typrelid,
 													 fselect->fieldnum);
-				appendStringInfo(buf, ".%s", quote_identifier(fieldname));
-				ReleaseSysCache(typetup);
+				/*
+				 * If the argument is simple enough, we could emit
+				 * arg.fieldname, but most cases where FieldSelect is used
+				 * are *not* simple.  For now, always use the projection-
+				 * function syntax.
+				 */
+				appendStringInfo(buf, "%s(", quote_identifier(fieldname));
+				get_rule_expr(fselect->arg, context);
+				appendStringInfoChar(buf, ')');
 			}
 			break;
 
@@ -1919,6 +1951,60 @@ get_rule_expr(Node *node, deparse_context *context)
 			}
 			break;
 
+		case T_NullTest:
+			{
+				NullTest   *ntest = (NullTest *) node;
+
+				appendStringInfo(buf, "(");
+				get_rule_expr(ntest->arg, context);
+				switch (ntest->nulltesttype)
+				{
+					case IS_NULL:
+						appendStringInfo(buf, " IS NULL)");
+						break;
+					case IS_NOT_NULL:
+						appendStringInfo(buf, " IS NOT NULL)");
+						break;
+					default:
+						elog(ERROR, "get_rule_expr: unexpected nulltesttype %d",
+							 (int) ntest->nulltesttype);
+				}
+			}
+			break;
+
+		case T_BooleanTest:
+			{
+				BooleanTest *btest = (BooleanTest *) node;
+
+				appendStringInfo(buf, "(");
+				get_rule_expr(btest->arg, context);
+				switch (btest->booltesttype)
+				{
+					case IS_TRUE:
+						appendStringInfo(buf, " IS TRUE)");
+						break;
+					case IS_NOT_TRUE:
+						appendStringInfo(buf, " IS NOT TRUE)");
+						break;
+					case IS_FALSE:
+						appendStringInfo(buf, " IS FALSE)");
+						break;
+					case IS_NOT_FALSE:
+						appendStringInfo(buf, " IS NOT FALSE)");
+						break;
+					case IS_UNKNOWN:
+						appendStringInfo(buf, " IS UNKNOWN)");
+						break;
+					case IS_NOT_UNKNOWN:
+						appendStringInfo(buf, " IS NOT UNKNOWN)");
+						break;
+					default:
+						elog(ERROR, "get_rule_expr: unexpected booltesttype %d",
+							 (int) btest->booltesttype);
+				}
+			}
+			break;
+
 		case T_SubLink:
 			get_sublink_expr(node, context);
 			break;
@@ -1949,25 +2035,6 @@ get_func_expr(Expr *expr, deparse_context *context)
 	List	   *l;
 	char	   *sep;
 
-	/*
-	 * nullvalue() and nonnullvalue() should get turned into special
-	 * syntax
-	 */
-	if (funcoid == F_NULLVALUE)
-	{
-		appendStringInfoChar(buf, '(');
-		get_rule_expr((Node *) lfirst(expr->args), context);
-		appendStringInfo(buf, " ISNULL)");
-		return;
-	}
-	if (funcoid == F_NONNULLVALUE)
-	{
-		appendStringInfoChar(buf, '(');
-		get_rule_expr((Node *) lfirst(expr->args), context);
-		appendStringInfo(buf, " NOTNULL)");
-		return;
-	}
-
 	/*
 	 * Get the functions pg_proc tuple
 	 */
@@ -1987,49 +2054,30 @@ get_func_expr(Expr *expr, deparse_context *context)
 	if (exprIsLengthCoercion((Node *) expr, &coercedTypmod))
 	{
 		Node	   *arg = lfirst(expr->args);
+		char	   *typdesc;
 
 		/*
-		 * Strip off any RelabelType on the input, so we don't print
-		 * redundancies like x::bpchar::char(8).
+		 * Strip off any type coercions on the input, so we don't print
+		 * redundancies like x::bpchar::character(8).
 		 *
 		 * XXX Are there any cases where this is a bad idea?
 		 */
-		if (IsA(arg, RelabelType))
-			arg = ((RelabelType *) arg)->arg;
+		arg = strip_type_coercion(arg, procStruct->prorettype);
+
 		appendStringInfoChar(buf, '(');
 		get_rule_expr(arg, context);
-		appendStringInfo(buf, ")::");
 
 		/*
 		 * Show typename with appropriate length decoration. Note that
-		 * since exprIsLengthCoercion succeeded, the function name is the
-		 * same as its output type name.
+		 * since exprIsLengthCoercion succeeded, the function's output
+		 * type is the right thing to report.  Also note we don't need
+		 * to quote the result of format_type_with_typemod: it takes
+		 * care of double-quoting any identifier that needs it.
 		 */
-		if (strcmp(proname, "bpchar") == 0)
-		{
-			if (coercedTypmod > (int32) VARHDRSZ)
-				appendStringInfo(buf, "char(%d)", coercedTypmod - VARHDRSZ);
-			else
-				appendStringInfo(buf, "char");
-		}
-		else if (strcmp(proname, "varchar") == 0)
-		{
-			if (coercedTypmod > (int32) VARHDRSZ)
-				appendStringInfo(buf, "varchar(%d)", coercedTypmod - VARHDRSZ);
-			else
-				appendStringInfo(buf, "varchar");
-		}
-		else if (strcmp(proname, "numeric") == 0)
-		{
-			if (coercedTypmod >= (int32) VARHDRSZ)
-				appendStringInfo(buf, "numeric(%d,%d)",
-							 ((coercedTypmod - VARHDRSZ) >> 16) & 0xffff,
-								 (coercedTypmod - VARHDRSZ) & 0xffff);
-			else
-				appendStringInfo(buf, "numeric");
-		}
-		else
-			appendStringInfo(buf, "%s", quote_identifier(proname));
+		typdesc = format_type_with_typemod(procStruct->prorettype,
+										   coercedTypmod);
+		appendStringInfo(buf, ")::%s", typdesc);
+		pfree(typdesc);
 
 		ReleaseSysCache(proctup);
 		return;
@@ -2052,6 +2100,79 @@ get_func_expr(Expr *expr, deparse_context *context)
 }
 
 
+/*
+ * strip_type_coercion
+ *		Strip any type coercions at the top of the given expression tree,
+ *		as long as they are coercions to the given datatype.
+ *
+ * A RelabelType node is always a type coercion.  A function call is also
+ * considered a type coercion if it has one argument and the function name
+ * is the same as the (internal) name of its result type.
+ *
+ * XXX It'd be better if the parsetree retained some explicit indication
+ * of the coercion, so we didn't need these heuristics.
+ */
+static Node *
+strip_type_coercion(Node *expr, Oid resultType)
+{
+	if (expr == NULL || exprType(expr) != resultType)
+		return expr;
+
+	if (IsA(expr, RelabelType))
+		return strip_type_coercion(((RelabelType *) expr)->arg, resultType);
+
+	if (IsA(expr, Expr) &&((Expr *) expr)->opType == FUNC_EXPR)
+	{
+		Func	   *func;
+		HeapTuple	procTuple;
+		HeapTuple	typeTuple;
+		Form_pg_proc procStruct;
+		Form_pg_type typeStruct;
+
+		func = (Func *) (((Expr *) expr)->oper);
+		Assert(IsA(func, Func));
+		if (length(((Expr *) expr)->args) != 1)
+			return expr;
+		/* Lookup the function in pg_proc */
+		procTuple = SearchSysCache(PROCOID,
+								   ObjectIdGetDatum(func->funcid),
+								   0, 0, 0);
+		if (!HeapTupleIsValid(procTuple))
+			elog(ERROR, "cache lookup for proc %u failed", func->funcid);
+		procStruct = (Form_pg_proc) GETSTRUCT(procTuple);
+		/* Double-check func has one arg and correct result type */
+		if (procStruct->pronargs != 1 ||
+			procStruct->prorettype != resultType)
+		{
+			ReleaseSysCache(procTuple);
+			return expr;
+		}
+		/* See if function has same name as its result type */
+		typeTuple = SearchSysCache(TYPEOID,
+								ObjectIdGetDatum(procStruct->prorettype),
+								   0, 0, 0);
+		if (!HeapTupleIsValid(typeTuple))
+			elog(ERROR, "cache lookup for type %u failed",
+				 procStruct->prorettype);
+		typeStruct = (Form_pg_type) GETSTRUCT(typeTuple);
+		if (strncmp(NameStr(procStruct->proname),
+					NameStr(typeStruct->typname),
+					NAMEDATALEN) != 0)
+		{
+			ReleaseSysCache(procTuple);
+			ReleaseSysCache(typeTuple);
+			return expr;
+		}
+		/* Okay, it is indeed a type-coercion function */
+		ReleaseSysCache(procTuple);
+		ReleaseSysCache(typeTuple);
+		return strip_type_coercion(lfirst(((Expr *) expr)->args), resultType);
+	}
+
+	return expr;
+}
+
+
 /* ----------
  * get_tle_expr
  *
@@ -2114,7 +2235,6 @@ get_const_expr(Const *constval, deparse_context *context)
 
 	if (constval->constisnull)
 	{
-
 		/*
 		 * Always label the type of a NULL constant.  This not only
 		 * prevents misdecisions about the type, but it ensures that our
@@ -2453,6 +2573,34 @@ get_from_clause_item(Node *jtnode, Query *query, deparse_context *context)
 	dpns->namespace = sv_namespace;
 }
 
+/* ----------
+ * get_opclass_name			- fetch name of an index operator class
+ *
+ * The opclass name is appended (after a space) to buf.
+ *
+ * Output is suppressed if the opclass is the default for the given
+ * actual_datatype.  (If you don't want this behavior, just pass
+ * InvalidOid for actual_datatype.)
+ * ----------
+ */
+static void
+get_opclass_name(Oid opclass, Oid actual_datatype,
+				 StringInfo buf)
+{
+	HeapTuple	ht_opc;
+	Form_pg_opclass opcrec;
+
+	ht_opc = SearchSysCache(CLAOID,
+							ObjectIdGetDatum(opclass),
+							0, 0, 0);
+	if (!HeapTupleIsValid(ht_opc))
+		elog(ERROR, "cache lookup failed for opclass %u", opclass);
+	opcrec = (Form_pg_opclass) GETSTRUCT(ht_opc);
+	if (actual_datatype != opcrec->opcintype || !opcrec->opcdefault)
+		appendStringInfo(buf, " %s",
+						 quote_identifier(NameStr(opcrec->opcname)));
+	ReleaseSysCache(ht_opc);
+}
 
 /* ----------
  * tleIsArrayAssign			- check for array assignment
@@ -2490,7 +2638,6 @@ tleIsArrayAssign(TargetEntry *tle)
 static char *
 quote_identifier(char *ident)
 {
-
 	/*
 	 * Can avoid quoting if ident starts with a lowercase letter and
 	 * contains only lowercase letters, digits, and underscores, *and* is
@@ -2522,7 +2669,6 @@ quote_identifier(char *ident)
 
 	if (safe)
 	{
-
 		/*
 		 * Check for keyword.  This test is overly strong, since many of
 		 * the "keywords" known to the parser are usable as column names,
