@@ -3,7 +3,7 @@
  *
  * Copyright 2000 by PostgreSQL Global Development Group
  *
- * $Header: /home/projects/pgsql/cvsroot/pgsql/src/bin/psql/command.c,v 1.49.2.1 2001/05/06 21:44:32 petere Exp $
+ * $Header: /cvsroot/pgsql-server/src/bin/psql/command.c,v 1.64.2.1 2002/02/25 21:37:47 tgl Exp $
  */
 #include "postgres_fe.h"
 #include "command.h"
@@ -11,6 +11,9 @@
 #include <errno.h>
 #include <assert.h>
 #include <ctype.h>
+#ifdef HAVE_PWD_H
+#include <pwd.h>
+#endif
 #ifndef WIN32
 #include <sys/types.h>			/* for umask() */
 #include <sys/stat.h>			/* for stat() */
@@ -51,11 +54,17 @@ static backslashResult exec_command(const char *cmd,
 			 const char **continue_parse,
 			 PQExpBuffer query_buf);
 
+/* different ways for scan_option to handle parameter words */
 enum option_type
 {
-	OT_NORMAL, OT_SQLID, OT_FILEPIPE
+	OT_NORMAL,					/* normal case */
+	OT_SQLID,					/* treat as SQL identifier */
+	OT_SQLIDHACK,				/* SQL identifier, but don't downcase */
+	OT_FILEPIPE					/* it's a file or pipe */
 };
-static char *scan_option(char **string, enum option_type type, char *quote);
+
+static char *scan_option(char **string, enum option_type type,
+						 char *quote, bool semicolon);
 static char *unescape(const unsigned char *source, size_t len);
 
 static bool do_edit(const char *filename_arg, PQExpBuffer query_buf);
@@ -127,33 +136,35 @@ HandleSlashCmds(const char *line,
 
 	if (status == CMD_UNKNOWN)
 	{
-
 		/*
-		 * If the command was not recognized, try inserting a space after
-		 * the first letter and call again. The one letter commands allow
-		 * arguments to start immediately after the command, but that is
-		 * no longer encouraged.
+		 * If the command was not recognized, try to parse it as a
+		 * one-letter command with immediately following argument (a
+		 * still-supported, but no longer encouraged, syntax).
 		 */
 		char		new_cmd[2];
 
 		new_cmd[0] = my_line[0];
 		new_cmd[1] = '\0';
 
+		/* use line for options, because my_line was clobbered above */
 		status = exec_command(new_cmd, line + 1, &continue_parse, query_buf);
 
-		/* continue_parse must be relative to my_line for calculation below */
+		/*
+		 * continue_parse must be relative to my_line for calculation
+		 * below
+		 */
 		continue_parse += my_line - line;
 
 #if 0							/* turned out to be too annoying */
 		if (status != CMD_UNKNOWN && isalpha((unsigned char) new_cmd[0]))
-			psql_error("Warning: this syntax is deprecated\n");
+			psql_error("Warning: This syntax is deprecated.\n");
 #endif
 	}
 
 	if (status == CMD_UNKNOWN)
 	{
 		if (pset.cur_cmd_interactive)
-			fprintf(stderr, "Invalid command \\%s. Try \\? for help.\n", my_line);
+			fprintf(stderr, gettext("Invalid command \\%s. Try \\? for help.\n"), my_line);
 		else
 			psql_error("invalid command \\%s\n", my_line);
 		status = CMD_ERROR;
@@ -216,7 +227,7 @@ exec_command(const char *cmd,
 	/* \C -- override table title (formerly change HTML caption) */
 	else if (strcmp(cmd, "C") == 0)
 	{
-		char	   *opt = scan_option(&string, OT_NORMAL, NULL);
+		char	   *opt = scan_option(&string, OT_NORMAL, NULL, true);
 
 		success = do_pset("title", opt, &pset.popt, quiet);
 		free(opt);
@@ -238,8 +249,17 @@ exec_command(const char *cmd,
 		char		opt1q,
 					opt2q;
 
-		opt1 = scan_option(&string, OT_NORMAL, &opt1q);
-		opt2 = scan_option(&string, OT_NORMAL, &opt2q);
+		/*
+		 * Ideally we should treat the arguments as SQL identifiers.  But for
+		 * backwards compatibility with 7.2 and older pg_dump files, we have
+		 * to take unquoted arguments verbatim (don't downcase them).
+		 * For now, double-quoted arguments may be stripped of double quotes
+		 * (as if SQL identifiers).  By 7.4 or so, pg_dump files can be
+		 * expected to double-quote all mixed-case \connect arguments,
+		 * and then we can get rid of OT_SQLIDHACK.
+		 */
+		opt1 = scan_option(&string, OT_SQLIDHACK, &opt1q, true);
+		opt2 = scan_option(&string, OT_SQLIDHACK, &opt2q, true);
 
 		if (opt2)
 			/* gave username */
@@ -256,6 +276,47 @@ exec_command(const char *cmd,
 		free(opt2);
 	}
 
+	/* \cd */
+	else if (strcmp(cmd, "cd") == 0)
+	{
+		char	   *opt = scan_option(&string, OT_NORMAL, NULL, true);
+		char	   *dir;
+
+		if (opt)
+			dir = opt;
+		else
+		{
+#ifndef WIN32
+			struct passwd *pw;
+
+			pw = getpwuid(geteuid());
+			if (!pw)
+			{
+				psql_error("could not get home directory: %s\n", strerror(errno));
+				exit(EXIT_FAILURE);
+			}
+			dir = pw->pw_dir;
+#else							/* WIN32 */
+
+			/*
+			 * On Windows, 'cd' without arguments prints the current
+			 * directory, so if someone wants to code this here instead...
+			 */
+			dir = "/";
+#endif   /* WIN32 */
+		}
+
+		if (chdir(dir) == -1)
+		{
+			psql_error("\\%s: could not change directory to '%s': %s\n",
+					   cmd, dir, strerror(errno));
+			success = false;
+		}
+
+		if (opt)
+			free(opt);
+	}
+
 	/* \copy */
 	else if (strcasecmp(cmd, "copy") == 0)
 	{
@@ -274,7 +335,8 @@ exec_command(const char *cmd,
 		char	   *name;
 		bool		show_verbose;
 
-		name = scan_option(&string, OT_SQLID, NULL);
+		name = scan_option(&string, OT_SQLID, NULL, true);
+
 		show_verbose = strchr(cmd, '+') ? true : false;
 
 		switch (cmd[1])
@@ -315,6 +377,9 @@ exec_command(const char *cmd,
 			case 'S':
 				success = listTables(&cmd[1], name, show_verbose);
 				break;
+			case 'u':
+				success = describeUsers(name);
+				break;
 			default:
 				status = CMD_UNKNOWN;
 		}
@@ -337,7 +402,7 @@ exec_command(const char *cmd,
 		}
 		else
 		{
-			fname = scan_option(&string, OT_NORMAL, NULL);
+			fname = scan_option(&string, OT_NORMAL, NULL, true);
 			status = do_edit(fname, query_buf) ? CMD_NEWEDIT : CMD_ERROR;
 			free(fname);
 		}
@@ -357,7 +422,7 @@ exec_command(const char *cmd,
 		else
 			fout = stdout;
 
-		while ((value = scan_option(&string, OT_NORMAL, &quoted)))
+		while ((value = scan_option(&string, OT_NORMAL, &quoted, false)))
 		{
 			if (!quoted && strcmp(value, "-n") == 0)
 				no_newline = true;
@@ -378,7 +443,7 @@ exec_command(const char *cmd,
 	/* \encoding -- set/show client side encoding */
 	else if (strcmp(cmd, "encoding") == 0)
 	{
-		char	   *encoding = scan_option(&string, OT_NORMAL, NULL);
+		char	   *encoding = scan_option(&string, OT_NORMAL, NULL, false);
 
 		if (!encoding)
 			/* show encoding */
@@ -397,7 +462,7 @@ exec_command(const char *cmd,
 				SetVariable(pset.vars, "ENCODING", pg_encoding_to_char(pset.encoding));
 			}
 #else
-			psql_error("\\%s: multi-byte support is not enabled\n", cmd);
+			psql_error("\\%s: multibyte support is not enabled\n", cmd);
 #endif
 			free(encoding);
 		}
@@ -406,7 +471,7 @@ exec_command(const char *cmd,
 	/* \f -- change field separator */
 	else if (strcmp(cmd, "f") == 0)
 	{
-		char	   *fname = scan_option(&string, OT_NORMAL, NULL);
+		char	   *fname = scan_option(&string, OT_NORMAL, NULL, false);
 
 		success = do_pset("fieldsep", fname, &pset.popt, quiet);
 		free(fname);
@@ -415,7 +480,7 @@ exec_command(const char *cmd,
 	/* \g means send query */
 	else if (strcmp(cmd, "g") == 0)
 	{
-		char	   *fname = scan_option(&string, OT_FILEPIPE, NULL);
+		char	   *fname = scan_option(&string, OT_FILEPIPE, NULL, false);
 
 		if (!fname)
 			pset.gfname = NULL;
@@ -447,7 +512,7 @@ exec_command(const char *cmd,
 	/* \i is include file */
 	else if (strcmp(cmd, "i") == 0 || strcmp(cmd, "include") == 0)
 	{
-		char	   *fname = scan_option(&string, OT_NORMAL, NULL);
+		char	   *fname = scan_option(&string, OT_NORMAL, NULL, true);
 
 		if (!fname)
 		{
@@ -475,8 +540,8 @@ exec_command(const char *cmd,
 		char	   *opt1,
 				   *opt2;
 
-		opt1 = scan_option(&string, OT_NORMAL, NULL);
-		opt2 = scan_option(&string, OT_NORMAL, NULL);
+		opt1 = scan_option(&string, OT_NORMAL, NULL, true);
+		opt2 = scan_option(&string, OT_NORMAL, NULL, true);
 
 		if (strcmp(cmd + 3, "export") == 0)
 		{
@@ -525,7 +590,7 @@ exec_command(const char *cmd,
 	/* \o -- set query output */
 	else if (strcmp(cmd, "o") == 0 || strcmp(cmd, "out") == 0)
 	{
-		char	   *fname = scan_option(&string, OT_FILEPIPE, NULL);
+		char	   *fname = scan_option(&string, OT_FILEPIPE, NULL, true);
 
 		success = setQFout(fname);
 		free(fname);
@@ -537,15 +602,15 @@ exec_command(const char *cmd,
 		if (query_buf && query_buf->len > 0)
 			puts(query_buf->data);
 		else if (!quiet)
-			puts("Query buffer is empty.");
+			puts(gettext("Query buffer is empty."));
 		fflush(stdout);
 	}
 
 	/* \pset -- set printing parameters */
 	else if (strcmp(cmd, "pset") == 0)
 	{
-		char	   *opt0 = scan_option(&string, OT_NORMAL, NULL);
-		char	   *opt1 = scan_option(&string, OT_NORMAL, NULL);
+		char	   *opt0 = scan_option(&string, OT_NORMAL, NULL, false);
+		char	   *opt1 = scan_option(&string, OT_NORMAL, NULL, false);
 
 		if (!opt0)
 		{
@@ -568,25 +633,25 @@ exec_command(const char *cmd,
 	{
 		resetPQExpBuffer(query_buf);
 		if (!quiet)
-			puts("Query buffer reset (cleared).");
+			puts(gettext("Query buffer reset (cleared)."));
 	}
 
 	/* \s save history in a file or show it on the screen */
 	else if (strcmp(cmd, "s") == 0)
 	{
-		char	   *fname = scan_option(&string, OT_NORMAL, NULL);
+		char	   *fname = scan_option(&string, OT_NORMAL, NULL, true);
 
 		success = saveHistory(fname ? fname : "/dev/tty");
 
 		if (success && !quiet && fname)
-			printf("Wrote history to %s.\n", fname);
+			printf(gettext("Wrote history to %s.\n"), fname);
 		free(fname);
 	}
 
 	/* \set -- generalized set variable/option command */
 	else if (strcmp(cmd, "set") == 0)
 	{
-		char	   *opt0 = scan_option(&string, OT_NORMAL, NULL);
+		char	   *opt0 = scan_option(&string, OT_NORMAL, NULL, false);
 
 		if (!opt0)
 		{
@@ -604,18 +669,17 @@ exec_command(const char *cmd,
 		}
 		else
 		{
-
 			/*
 			 * Set variable to the concatenation of the arguments.
 			 */
 			char	   *newval = NULL;
 			char	   *opt;
 
-			opt = scan_option(&string, OT_NORMAL, NULL);
+			opt = scan_option(&string, OT_NORMAL, NULL, false);
 			newval = xstrdup(opt ? opt : "");
 			free(opt);
 
-			while ((opt = scan_option(&string, OT_NORMAL, NULL)))
+			while ((opt = scan_option(&string, OT_NORMAL, NULL, false)))
 			{
 				newval = realloc(newval, strlen(newval) + strlen(opt) + 1);
 				if (!newval)
@@ -645,7 +709,7 @@ exec_command(const char *cmd,
 	/* \T -- define html <table ...> attributes */
 	else if (strcmp(cmd, "T") == 0)
 	{
-		char	   *value = scan_option(&string, OT_NORMAL, NULL);
+		char	   *value = scan_option(&string, OT_NORMAL, NULL, false);
 
 		success = do_pset("tableattr", value, &pset.popt, quiet);
 		free(value);
@@ -654,14 +718,14 @@ exec_command(const char *cmd,
 	/* \unset */
 	else if (strcmp(cmd, "unset") == 0)
 	{
-		char	   *opt = scan_option(&string, OT_NORMAL, NULL);
+		char	   *opt = scan_option(&string, OT_NORMAL, NULL, false);
 
 		if (!opt)
 		{
 			psql_error("\\%s: missing required argument\n", cmd);
 			success = false;
 		}
-		if (!SetVariable(pset.vars, opt, NULL))
+		else if (!SetVariable(pset.vars, opt, NULL))
 		{
 			psql_error("\\%s: error\n", cmd);
 			success = false;
@@ -683,7 +747,7 @@ exec_command(const char *cmd,
 		}
 		else
 		{
-			fname = scan_option(&string, OT_FILEPIPE, NULL);
+			fname = scan_option(&string, OT_FILEPIPE, NULL, true);
 
 			if (!fname)
 			{
@@ -738,7 +802,7 @@ exec_command(const char *cmd,
 	/* \z -- list table rights (grant/revoke) */
 	else if (strcmp(cmd, "z") == 0)
 	{
-		char	   *opt = scan_option(&string, OT_SQLID, NULL);
+		char	   *opt = scan_option(&string, OT_SQLID, NULL, true);
 
 		success = permissionsList(opt);
 		free(opt);
@@ -769,7 +833,7 @@ exec_command(const char *cmd,
 		char	   *value;
 
 		fprintf(stderr, "+ optstr = |%s|\n", options_string);
-		while ((value = scan_option(&string, OT_NORMAL, NULL)))
+		while ((value = scan_option(&string, OT_NORMAL, NULL, true)))
 		{
 			fprintf(stderr, "+ opt(%d) = |%s|\n", i++, value);
 			free(value);
@@ -784,7 +848,7 @@ exec_command(const char *cmd,
 		status = CMD_ERROR;
 
 	/* eat the rest of the options string */
-	while ((val = scan_option(&string, OT_NORMAL, NULL)))
+	while ((val = scan_option(&string, OT_NORMAL, NULL, false)))
 	{
 		if (status != CMD_UNKNOWN)
 			psql_error("\\%s: extra argument '%s' ignored\n", cmd, val);
@@ -803,7 +867,7 @@ exec_command(const char *cmd,
  * scan_option()
  */
 static char *
-scan_option(char **string, enum option_type type, char *quote)
+scan_option(char **string, enum option_type type, char *quote, bool semicolon)
 {
 	unsigned int pos = 0;
 	char	   *options_string;
@@ -821,7 +885,6 @@ scan_option(char **string, enum option_type type, char *quote)
 
 	switch (options_string[pos])
 	{
-
 			/*
 			 * Double quoted string
 			 */
@@ -844,7 +907,7 @@ scan_option(char **string, enum option_type type, char *quote)
 
 				if (options_string[jj] == 0)
 				{
-					psql_error("parse error at end of line\n");
+					psql_error("parse error at the end of line\n");
 					*string = &options_string[jj];
 					return NULL;
 				}
@@ -860,7 +923,8 @@ scan_option(char **string, enum option_type type, char *quote)
 				 * If this is expected to be an SQL identifier like option
 				 * then we strip out the double quotes
 				 */
-				if (type == OT_SQLID)
+
+				if (type == OT_SQLID || type == OT_SQLIDHACK)
 				{
 					unsigned int k,
 								cc;
@@ -881,7 +945,6 @@ scan_option(char **string, enum option_type type, char *quote)
 					}
 					return_val[cc] = '\0';
 				}
-
 				else
 				{
 					strncpy(return_val, &options_string[pos], jj - pos + 1);
@@ -918,7 +981,7 @@ scan_option(char **string, enum option_type type, char *quote)
 
 				if (options_string[jj] == 0)
 				{
-					psql_error("parse error at end of line\n");
+					psql_error("parse error at the end of line\n");
 					*string = &options_string[jj];
 					return NULL;
 				}
@@ -946,7 +1009,7 @@ scan_option(char **string, enum option_type type, char *quote)
 				len = strcspn(options_string + pos + 1, "`");
 				if (options_string[pos + 1 + len] == 0)
 				{
-					psql_error("parse error at end of line\n");
+					psql_error("parse error at the end of line\n");
 					*string = &options_string[pos + 1 + len];
 					return NULL;
 				}
@@ -1072,6 +1135,16 @@ scan_option(char **string, enum option_type type, char *quote)
 				strncpy(return_val, &options_string[pos], token_end);
 				return_val[token_end] = 0;
 
+				/* Strip any trailing semi-colons for some types */
+				if (semicolon)
+				{
+					int			i;
+
+					for (i = strlen(return_val) - 1; i && return_val[i] == ';'; i--);
+					if (i < strlen(return_val) - 1)
+						return_val[i + 1] = '\0';
+				}
+
 				if (type == OT_SQLID)
 					for (cp = return_val; *cp; cp += PQmblen(cp, pset.encoding))
 						if (isupper((unsigned char) *cp))
@@ -1227,14 +1300,14 @@ do_connect(const char *new_dbname, const char *new_user)
 
 	/* need to prompt for password? */
 	if (pset.getPassword)
-		pwparam = prompted_password = simple_prompt("Password: ", 100, false);	/* need to save for
-																				 * free() */
+		pwparam = prompted_password = simple_prompt("Password: ", 100, false);
 
 	/*
-	 * Use old password if no new one given (if you didn't have an old
-	 * one, fine)
+	 * Use old password (if any) if no new one given and we are
+	 * reconnecting as same user
 	 */
-	if (!pwparam && oldconn)
+	if (!pwparam && oldconn && PQuser(oldconn) && userparam &&
+		strcmp(PQuser(oldconn), userparam) == 0)
 		pwparam = PQpass(oldconn);
 
 	do
@@ -1269,7 +1342,7 @@ do_connect(const char *new_dbname, const char *new_user)
 			PQfinish(pset.db);
 			if (oldconn)
 			{
-				fputs("Previous connection kept\n", stderr);
+				fputs(gettext("Previous connection kept\n"), stderr);
 				pset.db = oldconn;
 			}
 			else
@@ -1277,7 +1350,6 @@ do_connect(const char *new_dbname, const char *new_user)
 		}
 		else
 		{
-
 			/*
 			 * we don't want unpredictable things to happen in scripting
 			 * mode
@@ -1294,12 +1366,12 @@ do_connect(const char *new_dbname, const char *new_user)
 		if (!QUIET())
 		{
 			if (userparam != new_user)	/* no new user */
-				printf("You are now connected to database %s.\n", dbparam);
+				printf(gettext("You are now connected to database %s.\n"), dbparam);
 			else if (dbparam != new_dbname)		/* no new db */
-				printf("You are now connected as new user %s.\n", new_user);
+				printf(gettext("You are now connected as new user %s.\n"), new_user);
 			else
 /* both new */
-				printf("You are now connected to database %s as user %s.\n",
+				printf(gettext("You are now connected to database %s as user %s.\n"),
 					   PQdb(pset.db), PQuser(pset.db));
 		}
 
@@ -1412,7 +1484,6 @@ do_edit(const char *filename_arg, PQExpBuffer query_buf)
 #ifndef WIN32
 	struct stat before,
 				after;
-
 #endif
 
 	if (filename_arg)
@@ -1438,7 +1509,7 @@ do_edit(const char *filename_arg, PQExpBuffer query_buf)
 
 		if (fd == -1 || !stream)
 		{
-			psql_error("couldn't open temp file %s: %s\n", fname, strerror(errno));
+			psql_error("could not open temporary file %s: %s\n", fname, strerror(errno));
 			error = true;
 		}
 		else
@@ -1572,7 +1643,7 @@ _align2string(enum printFormat in)
 {
 	switch (in)
 	{
-			case PRINT_NOTHING:
+		case PRINT_NOTHING:
 			return "nothing";
 			break;
 		case PRINT_UNALIGNED:
@@ -1627,7 +1698,7 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 		}
 
 		if (!quiet)
-			printf("Output format is %s.\n", _align2string(popt->topt.format));
+			printf(gettext("Output format is %s.\n"), _align2string(popt->topt.format));
 	}
 
 	/* set border style/width */
@@ -1637,7 +1708,7 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 			popt->topt.border = atoi(value);
 
 		if (!quiet)
-			printf("Border style is %d.\n", popt->topt.border);
+			printf(gettext("Border style is %d.\n"), popt->topt.border);
 	}
 
 	/* set expanded/vertical mode */
@@ -1645,7 +1716,9 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 	{
 		popt->topt.expanded = !popt->topt.expanded;
 		if (!quiet)
-			printf("Expanded display is %s.\n", popt->topt.expanded ? "on" : "off");
+			printf(popt->topt.expanded
+				   ? gettext("Expanded display is on.\n")
+				   : gettext("Expanded display is off.\n"));
 	}
 
 	/* null display */
@@ -1657,7 +1730,7 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 			popt->nullPrint = xstrdup(value);
 		}
 		if (!quiet)
-			printf("Null display is '%s'.\n", popt->nullPrint ? popt->nullPrint : "");
+			printf(gettext("Null display is '%s'.\n"), popt->nullPrint ? popt->nullPrint : "");
 	}
 
 	/* field separator for unaligned text */
@@ -1669,7 +1742,7 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 			popt->topt.fieldSep = xstrdup(value);
 		}
 		if (!quiet)
-			printf("Field separator is '%s'.\n", popt->topt.fieldSep);
+			printf(gettext("Field separator is '%s'.\n"), popt->topt.fieldSep);
 	}
 
 	/* record separator for unaligned text */
@@ -1683,9 +1756,9 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 		if (!quiet)
 		{
 			if (strcmp(popt->topt.recordSep, "\n") == 0)
-				printf("Record separator is <newline>.");
+				printf(gettext("Record separator is <newline>."));
 			else
-				printf("Record separator is '%s'.\n", popt->topt.recordSep);
+				printf(gettext("Record separator is '%s'.\n"), popt->topt.recordSep);
 		}
 	}
 
@@ -1696,9 +1769,9 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 		if (!quiet)
 		{
 			if (popt->topt.tuples_only)
-				puts("Showing only tuples.");
+				puts(gettext("Showing only tuples."));
 			else
-				puts("Tuples only is off.");
+				puts(gettext("Tuples only is off."));
 		}
 	}
 
@@ -1714,9 +1787,9 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 		if (!quiet)
 		{
 			if (popt->title)
-				printf("Title is \"%s\".\n", popt->title);
+				printf(gettext("Title is \"%s\".\n"), popt->title);
 			else
-				printf("Title is unset.\n");
+				printf(gettext("Title is unset.\n"));
 		}
 	}
 
@@ -1732,9 +1805,9 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 		if (!quiet)
 		{
 			if (popt->topt.tableAttr)
-				printf("Table attribute is \"%s\".\n", popt->topt.tableAttr);
+				printf(gettext("Table attribute is \"%s\".\n"), popt->topt.tableAttr);
 			else
-				printf("Table attributes unset.\n");
+				printf(gettext("Table attributes unset.\n"));
 		}
 	}
 
@@ -1745,12 +1818,24 @@ do_pset(const char *param, const char *value, printQueryOpt *popt, bool quiet)
 		if (!quiet)
 		{
 			if (popt->topt.pager)
-				puts("Using pager is on.");
+				puts(gettext("Using pager is on."));
 			else
-				puts("Using pager is off.");
+				puts(gettext("Using pager is off."));
 		}
 	}
 
+	/* disable "(x rows)" footer */
+	else if (strcmp(param, "footer") == 0)
+	{
+		popt->default_footer = !popt->default_footer;
+		if (!quiet)
+		{
+			if (popt->default_footer)
+				puts(gettext("Default footer is on."));
+			else
+				puts(gettext("Default footer is off."));
+		}
+	}
 
 	else
 	{
